/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.examples.onlineSim.resilience;

import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.interfaces.simulation.IEventProcessor;
import com.net2plan.interfaces.simulation.SimAction;
import com.net2plan.interfaces.simulation.SimEvent;
import com.net2plan.libraries.GraphUtils.JUNGUtils;
import com.net2plan.libraries.TrafficComputationEngine;
import com.net2plan.utils.Pair;
import com.net2plan.utils.Triple;
import edu.uci.ics.jung.graph.Graph;
import java.util.*;
import org.apache.commons.collections15.Transformer;

/**
 * <p>Algorithm which implements a generic protection mechanism using pre-defined 
 * protection segments in the network design.</p>
 * 
 * <p>Upon failure, try to route the traffic over the primary path if it is available 
 * and there is enough capacity. Otherwise, try to route the traffic over a backup path 
 * including some protection segment. To do this, an auxiliary graph composed of links 
 * in the original path and the associated protection segments with enough spare capacity 
 * is generated. Then, Dijkstra's shortest path algorithm is applied to get the backup path. 
 * If fails, the route cannot be restored.</p>
 * 
 * <p>Upon reparation, try to restore each rerouted connection to their original path, 
 * using a make-before-break strategy, only in case 'revertiveMode' is enabled. Then,
 * still-failing routes are tried to be restored as under failure events.</p>
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @version 1.1, May 2015
 */
public class NRSim_AA_genericProtectionAlgorithm extends IEventProcessor
{
	private double PRECISION_FACTOR;
	private boolean revertiveMode;
	
	@Override
	public String getDescription()
	{
		String NEWLINE = String.format("%n");
		StringBuilder description = new StringBuilder();
		description.append("Algorithm which implements a generic protection mechanism using pre-defined protection segments in the network design.").append(NEWLINE).append(NEWLINE);
		description.append("Upon failure, try to route the traffic over the primary path if it is available and there is enough capacity. Otherwise, try to route the traffic over a backup path including some protection segment. To do this, an auxiliary graph composed of links in the original path and the associated protection segments with enough spare capacity is generated. Then, Dijkstra's shortest path algorithm is applied to get the backup path. If fails, the route cannot be restored.").append(NEWLINE).append(NEWLINE);
		description.append("Upon reparation, try to restore each rerouted connection to their original path, using a make-before-break strategy, only in case 'revertiveMode' is enabled. Then, still-failing routes are tried to be restored as under failure events.");

		return description.toString();
	}

	@Override
	public List<Triple<String, String, String>> getParameters()
	{
		List<Triple<String, String, String>> parameters = new LinkedList<Triple<String, String, String>>();
		parameters.add(Triple.of("revertiveMode", "#boolean# true", "Indicates whether routes are restored to their primary paths when become available"));
		
		return parameters;
	}
	
	@Override
	public void initialize(NetPlan initialNetPlan, NetPlan currentNetPlan, Map<String, String> algorithmParameters, Map<String, String> simulationParameters, Map<String, String> net2planParameters)
	{
		int N = currentNetPlan.getNumberOfNodes();
		int E = currentNetPlan.getNumberOfLinks();
		int D = currentNetPlan.getNumberOfDemands();
		int R = currentNetPlan.getNumberOfRoutes();
		if (N == 0 || E == 0 || D == 0 || R == 0) throw new Net2PlanException("A physical topology (nodes and links), demands and routing are required");
		
		PRECISION_FACTOR = Double.parseDouble(net2planParameters.get("precisionFactor"));
		revertiveMode = Boolean.parseBoolean(algorithmParameters.get("revertiveMode"));
	}

	@Override
	public void processEvent(NetPlan currentNetPlan, SimEvent event)
	{
		/* Restore as many protection segments as possible */
		currentNetPlan.setAllProtectionSegmentsUp();

		/* Check event type */
		boolean isFailureEvent = false;
		boolean isReparationEvent = false;
		if (event.containsAny(EnumSet.of(SimAction.ActionType.NODE_DOWN, SimAction.ActionType.LINK_DOWN))) isFailureEvent = true;
		if (event.containsAny(EnumSet.of(SimAction.ActionType.NODE_UP, SimAction.ActionType.LINK_UP))) isReparationEvent = true;
		if (isFailureEvent && isReparationEvent) throw new Net2PlanException("Failures and reparations cannot be processed as a single event");
		if (!isFailureEvent && !isReparationEvent) throw new Net2PlanException("This algorithm only processes failure and reparation events");
		
		if (isFailureEvent)
		{
			/* Try to reroute as much failing traffic as possible */
			rerouteFailingTraffic(currentNetPlan);
		}
		else
		{
			/* If revertiveMode is applied, then try to restore to the original path only those up routes */
			if (revertiveMode) restoreToPrimary(currentNetPlan);
			
			/* Try to reroute as much failing traffic as possible */
			rerouteFailingTraffic(currentNetPlan);
		}
	}

	/**
	 * Applies the restoration mechanism to all failing routes.
	 * 
	 * @param currentNetPlan Current network state
	 * @since 1.0
	 */
	private void rerouteFailingTraffic(NetPlan currentNetPlan)
	{
		Set<Long> segmentsDown = currentNetPlan.getProtectionSegmentsDown();
			
		/* Retrieve the set of routes that now are down */
		Set<Long> routesDown = new LinkedHashSet<Long>(currentNetPlan.getRoutesDown());
		for(long routeId : routesDown)
		{
			long ingressNodeId = currentNetPlan.getRouteIngressNode(routeId);
			long egressNodeId = currentNetPlan.getRouteEgressNode(routeId);
			if (currentNetPlan.isNodeDown(ingressNodeId) || currentNetPlan.isNodeDown(egressNodeId)) continue;

			List<Long> seqLinks_primaryPath_thisRoute = currentNetPlan.getRouteOriginalSequenceOfLinks(routeId);
			double x_p_thisRoute = currentNetPlan.getRouteOriginalCarriedTraffic(routeId);

			/* First, check whether original path is active */
			if (currentNetPlan.isPathRestorable(seqLinks_primaryPath_thisRoute))
			{
				double allowed_x_p_primaryPath = Math.min(x_p_thisRoute, TrafficComputationEngine.getPathBottleneckCapacity(currentNetPlan, seqLinks_primaryPath_thisRoute));

				if (allowed_x_p_primaryPath > 0 && x_p_thisRoute < allowed_x_p_primaryPath + PRECISION_FACTOR)
				{
					currentNetPlan.setRouteSequenceOfLinks(routeId, seqLinks_primaryPath_thisRoute);
					currentNetPlan.setRouteUp(routeId);
					continue;
				}
			}

			/* Then, try to recover traffic by adding protection segments to the original path */

			/* Compute set of candidate segments: those associated to the route excepts those down */
			List<Long> candidateSegmentIds = new LinkedList<Long>(currentNetPlan.getRouteBackupSegmentList(routeId));
			candidateSegmentIds.removeAll(segmentsDown);

			Map<Long, Pair<Long, Long>> linkMap_thisNodePair = new LinkedHashMap<Long, Pair<Long, Long>>();
			Map<Long, Double> linkCapacityMap_thisNodePair = new LinkedHashMap<Long, Double>();
			for(long linkId : seqLinks_primaryPath_thisRoute)
			{
				if (currentNetPlan.isLinkDown(linkId)) continue;

				Pair<Long, Long> nodePair_thisLink = currentNetPlan.getLinkOriginDestinationNodePair(linkId);
				long originNodeId = nodePair_thisLink.getFirst();
				long destinationNodeId = nodePair_thisLink.getSecond();
				if (currentNetPlan.isNodeDown(originNodeId) || currentNetPlan.isNodeDown(destinationNodeId)) continue;

				linkMap_thisNodePair.put(linkId, nodePair_thisLink);
				linkCapacityMap_thisNodePair.put(linkId, currentNetPlan.getLinkSpareCapacity(linkId));
			}

			for(long segmentId : currentNetPlan.getRouteBackupSegmentList(routeId))
			{
				if (currentNetPlan.isProtectionSegmentDown(segmentId)) continue;

				long auxLinkId = -1 - segmentId;
				Pair<Long, Long> nodePair_thisSegment = currentNetPlan.getProtectionSegmentOriginDestinationNodePair(segmentId);
				linkMap_thisNodePair.put(auxLinkId, nodePair_thisSegment);
				linkCapacityMap_thisNodePair.put(auxLinkId, currentNetPlan.getProtectionSegmentSpareCapacity(segmentId));
			}

			Graph<Long, Long> graph_thisRoute = JUNGUtils.getGraphFromLinkMap(linkMap_thisNodePair);
			Transformer<Long, Double> capacityTransformer = JUNGUtils.getEdgeWeightTransformer(linkCapacityMap_thisNodePair);
			List<Long> seqLinks = JUNGUtils.getCapacitatedShortestPath(graph_thisRoute, null, ingressNodeId, egressNodeId, capacityTransformer, x_p_thisRoute);
			if (seqLinks.isEmpty()) continue;

			currentNetPlan.setRouteSequenceOfLinks(routeId, seqLinks);
			currentNetPlan.setRouteUp(routeId);
		}
	}
	
	/**
	 * Tries to restore routes using an alternate path to their original path, if revertive mode 
	 * is enabled.
	 * 
	 * @param currentNetPlan Current network state
	 * @since 1.0
	 */
	private void restoreToPrimary(NetPlan currentNetPlan)
	{
		for(long routeId : currentNetPlan.getRouteIds())
		{
			boolean isRouteDown = currentNetPlan.isRouteDown(routeId);
			if (isRouteDown) continue;

			/* Check whether primary path is available */
			List<Long> seqLinks_primaryPath_thisRoute = currentNetPlan.getRouteOriginalSequenceOfLinks(routeId);
			boolean primaryPathAvailable = currentNetPlan.isPathRestorable(seqLinks_primaryPath_thisRoute);
			if (!primaryPathAvailable) continue;

			/* Check whether route is using its primary path and carrying the original traffic, then skip it */
			List<Long> currentSeqLinks_thisRoute = new LinkedList<Long>(currentNetPlan.getRouteSequenceOfLinks(routeId));
			double x_p_primaryPath_thisRoute = currentNetPlan.getRouteOriginalCarriedTraffic(routeId);
			double current_x_p_thisRoute = currentNetPlan.getRouteCarriedTraffic(routeId);
			if (currentSeqLinks_thisRoute.equals(seqLinks_primaryPath_thisRoute) && x_p_primaryPath_thisRoute < current_x_p_thisRoute + PRECISION_FACTOR) continue;

			/* Try to restore route to its primary path */
			currentNetPlan.setRouteSequenceOfLinks(routeId, seqLinks_primaryPath_thisRoute);
			currentNetPlan.setRouteCarriedTraffic(routeId, x_p_primaryPath_thisRoute);

			/* If fails (some link is oversubscribed), leave it as was */
			if (currentNetPlan.getLinkMaximumUtilization() > 1 + PRECISION_FACTOR)
			{
				currentNetPlan.setRouteSequenceOfLinks(routeId, currentSeqLinks_thisRoute);
				currentNetPlan.setRouteCarriedTraffic(routeId, current_x_p_thisRoute);
			}
		}
	}
}